from __future__ import annotations

__version__ = "26.03.17"
__uri__ = "https://github.com/nichollsh/InferAGNI"
__author__ = "Harrison Nicholls"
__email__ = "harrison.nicholls@ast.cam.ac.uk"
__license__ = "GPLv3"

import inferagni.cli as cli
import inferagni.planets as planets

# import inferagni.grid as grid
# import inferagni.plot as plot
# import inferagni.retrieve as retrieve
# import inferagni.util as util

__all__ = ["cli", "planets"]  # "plot", "retrieve", "util",  "grid",
